@@ -882,7 +882,9 @@ undocumented(File, Objs, Options) -->
     { memberchk(module(Module), Options),
       memberchk(public(Exports), Options),
       select_undocumented(Exports, Module, Objs, Undoc),
-      re_exported_doc(Undoc, File, Module, REObjs, ReallyUnDoc)
+      re_exported_doc(Undoc, File, Module, UREObjs, ReallyUnDoc),
+      sort(2, @=<, UREObjs, REObjs) % UREObjs = doc(PI,Pos,Comment)
+                                    % i.e., sort on Pos
     },
     !,
     re_exported_doc(REObjs, Options),
@@ -2249,9 +2251,16 @@ pi(M:PI, Options) -->
     pi(PI, Options).
 pi(Name/Arity, _) -->
     !,
-    html([Name, /, Arity]).
+    html([Name, /, \arity(Arity)]).
 pi(Name//Arity, _) -->
-    html([Name, //, Arity]).
+    html([Name, //, \arity(Arity)]).
+
+arity(Arity) -->
+    { var(Arity) },
+    !,
+    html('_').
+arity(Arity) -->
+    html(Arity).
 
 pi_type(_:PI) -->
     !,
