@@ -3,8 +3,9 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2011-2015, University of Amsterdam
+    Copyright (c)  2011-2024, University of Amsterdam
                               VU University Amsterdam
+			      SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -35,6 +36,8 @@
 
 #include "pl-ctype.h"
 #include "pl-text.h"
+#include "pl-utf8.h"
+#include "pl-buffer.h"
 #include "../pl-read.h"
 #include "../pl-fli.h"
 #include <ctype.h>
@@ -54,8 +57,8 @@ See manual for details.
 
 typedef struct
 { atom_t	name;			/* name of the class */
-  int (*test)(wint_t chr);		/* boolean */
-  int (*reverse)(wint_t chr);		/* reverse mapping */
+  int (*test)(int chr);			/* boolean */
+  int (*reverse)(int chr);		/* reverse mapping */
   short		arity;			/* arity of class (i.e. lower('A')) */
   short		ctx_type;		/* CTX_* */
 } char_type;
@@ -73,69 +76,85 @@ typedef struct
 
 
 static int
-iswhite(wint_t chr)
+iswhite(int chr)
 { return chr == ' ' || chr == '\t';
 }
 
 
 static int
-fiscsym(wint_t chr)
+fiscsym(int chr)
 { return iswalnum(chr) || chr == '_';
 }
 
 
 static int
-fiscsymf(wint_t chr)
+fiscsymf(int chr)
 { return iswalpha(chr) || chr == '_';
 }
 
 static int
-iseof(wint_t chr)
-{ return chr == (wint_t)-1;
+iseof(int chr)
+{ return chr == -1;
 }
 
 static int
-iseol(wint_t chr)
+iseol(int chr)
 { return chr >= 10 && chr <= 13;
 }
 
 static int
-isnl(wint_t chr)
+isnl(int chr)
 { return chr == '\n';
 }
 
 static int
-isperiod(wint_t chr)
+isperiod(int chr)
 { return chr && strchr(".?!", chr) != NULL;
 }
 
 static int
-isquote(wint_t chr)
+isquote(int chr)
 { return chr && strchr("'`\"", chr) != NULL;
 }
 
 static int
-fupper(wint_t chr)
-{ return iswlower(chr) ? (int)towupper(chr) : -1;
+fupper(int chr)
+{
+#if SIZEOF_WINT_T == 2
+  if ( chr > 0xffff ) return -1;
+#endif
+  return iswlower(chr) ? (int)towupper(chr) : -1;
 }
 
 static int
-flower(wint_t chr)
-{ return iswupper(chr) ? (int)towlower(chr) : -1;
+flower(int chr)
+{
+#if SIZEOF_WINT_T == 2
+  if ( chr > 0xffff ) return -1;
+#endif
+  return iswupper(chr) ? (int)towlower(chr) : -1;
 }
 
 static int
-ftoupper(wint_t chr)
-{ return towupper(chr);
+ftoupper(int chr)
+{
+#if SIZEOF_WINT_T == 2
+  if ( chr > 0xffff ) return chr;
+#endif
+  return towupper(chr);
 }
 
 static int
-ftolower(wint_t chr)
-{ return towlower(chr);
+ftolower(int chr)
+{
+#if SIZEOF_WINT_T == 2
+  if ( chr > 0xffff ) return chr;
+#endif
+  return towlower(chr);
 }
 
 static int
-fparen(wint_t chr)
+fparen(int chr)
 { switch(chr)
   { case '(':
       return ')';
@@ -150,7 +169,7 @@ fparen(wint_t chr)
 
 
 static int
-rparen(wint_t chr)
+rparen(int chr)
 { switch(chr)
   { case ')':
       return '(';
@@ -165,7 +184,7 @@ rparen(wint_t chr)
 
 
 static int
-fdigit(wint_t chr)
+fdigit(int chr)
 { if ( chr <= 0xff && isdigit(chr) )
     return chr - '0';
   return -1;
@@ -173,7 +192,7 @@ fdigit(wint_t chr)
 
 
 static int
-rdigit(wint_t d)
+rdigit(int d)
 { if ( (int)d >= 0 && d <= 9 )
     return d+'0';
   return -1;
@@ -181,7 +200,15 @@ rdigit(wint_t d)
 
 
 static int
-fxdigit(wint_t chr)
+ff_decimal(int chr)
+{ if ( f_is_decimal(chr) )
+    return decimal_weight(chr);
+  else
+    return -1;
+}
+
+static int
+fxdigit(int chr)
 { if ( chr > 0xff )
     return -1;
   if ( isdigit(chr) )
@@ -195,7 +222,7 @@ fxdigit(wint_t chr)
 
 
 static int
-rxdigit(wint_t d)
+rxdigit(int d)
 { if ( (int)d >= 0 && d <= 9 )
     return d+'0';
   if ( d >= 10 && d <= 15 )
@@ -203,10 +230,14 @@ rxdigit(wint_t d)
   return -1;
 }
 
-
-
+#if SIZEOF_WINT_T == 2
 #define mkfunction(name) \
-	static int f ## name(wint_t chr) { return name(chr); }
+	static int f ## name(int chr) { if ( chr > 0xffff ) return 0; \
+					return name(chr); }
+#else
+#define mkfunction(name) \
+	static int f ## name(int chr) { return name(chr); }
+#endif
 
 mkfunction(iswalnum)
 mkfunction(iswalpha)
@@ -228,6 +259,8 @@ static const char_type char_types[] =
   { ATOM_prolog_var_start,	     f_is_prolog_var_start },
   { ATOM_prolog_atom_start,	     f_is_prolog_atom_start },
   { ATOM_prolog_identifier_continue, f_is_prolog_identifier_continue },
+  { ATOM_decimal,		     f_is_decimal },
+  { ATOM_decimal,		     ff_decimal, NULL, 1, CTX_CODE },
   { ATOM_prolog_symbol,		     f_is_prolog_symbol },
   { ATOM_csymf,			     fiscsymf },
   { ATOM_ascii,			     fisascii },
@@ -250,8 +283,8 @@ static const char_type char_types[] =
   { ATOM_to_lower,		     ftoupper,	ftolower, 1, CTX_CHAR },
   { ATOM_to_upper,		     ftolower,	ftoupper, 1, CTX_CHAR },
   { ATOM_paren,			     fparen,	rparen,   1, CTX_CHAR },
-  { ATOM_digit,			     fdigit,	rdigit,   1, CTX_CODE  },
-  { ATOM_xdigit,		     fxdigit,	rxdigit,  1, CTX_CODE  },
+  { ATOM_digit,			     fdigit,	rdigit,   1, CTX_CODE },
+  { ATOM_xdigit,		     fxdigit,	rxdigit,  1, CTX_CODE },
   { NULL_ATOM,			     NULL }
 };
 
@@ -272,7 +305,7 @@ char_type_by_name(atom_t name, int arity)
 static int
 advanceGen(generator *gen)
 { if ( gen->do_enum & ENUM_CHAR )
-  { if ( ++gen->current > PLMAXWCHAR )
+  { if ( ++gen->current > UNICODE_MAX )
       fail;
   } else
   { gen->class++;
@@ -330,8 +363,8 @@ do_char_type(term_t chr, term_t class, control_t h, int how)
 	return PL_error("char_type", 2, NULL, ERR_INSTANTIATION);
 
       if ( !(do_enum & ENUM_CHAR) )
-      { if ( !PL_get_char(chr, &c, TRUE) )
-	  fail;
+      { if ( !PL_get_char_ex(chr, &c, TRUE) )
+	  return FALSE;
 	if ( c == -1 )
 	  return PL_unify_atom(class, ATOM_end_of_file);
       }
@@ -345,9 +378,9 @@ do_char_type(term_t chr, term_t class, control_t h, int how)
 
       if ( do_enum == ENUM_NONE )
       { if ( arity == 0 )
-	  return (*cc->test)((wint_t)c) ? TRUE : FALSE;
+	  return (*cc->test)(c) ? TRUE : FALSE;
 	else
-	{ int rval = (*cc->test)((wint_t)c);
+	{ int rval = (*cc->test)(c);
 
 	  if ( rval >= 0 )
 	  { term_t a = PL_new_term_ref();
@@ -369,14 +402,14 @@ do_char_type(term_t chr, term_t class, control_t h, int how)
 	}
       }
 
-      if ( do_enum == ENUM_CHAR && arity == 1 )
+      if ( do_enum == ENUM_CHAR && arity == 1 && cc->reverse )
       {	term_t a = PL_new_term_ref();	/* char_type(X, lower('A')) */
 	int ca;
 
 	_PL_get_arg(1, class, a);
 	if ( !PL_is_variable(a) )
 	{ if ( PL_get_char(a, &ca, FALSE) )
-	  { int c = (*cc->reverse)((wint_t)ca);
+	  { int c = (*cc->reverse)(ca);
 
 	    if ( c < 0 )
 	      fail;
@@ -415,7 +448,7 @@ do_char_type(term_t chr, term_t class, control_t h, int how)
     goto error;
 
   for(;;)
-  { int rval = (*gen->class->test)((wint_t)gen->current);
+  { int rval = (*gen->class->test)(gen->current);
 
     if ( (gen->class->arity == 0 && rval) ||
 	 (gen->class->arity  > 0 && rval >= 0) )
@@ -487,40 +520,23 @@ PRED_IMPL("iswctype", 2, iswctype, 0)
 
 
 static int
-init_tout(PL_chars_t *t, size_t len)
-{ switch(t->encoding)
-  { case ENC_ISO_LATIN_1:
-      if ( len < sizeof(t->buf) )
-      { t->text.t = t->buf;
-	t->storage = PL_CHARS_LOCAL;
-      } else
-      { t->text.t = PL_malloc(len);
-	t->storage = PL_CHARS_MALLOC;
-      }
-      succeed;
-    case ENC_WCHAR:
-      if ( len*sizeof(pl_wchar_t) < sizeof(t->buf) )
-      { t->text.w = (pl_wchar_t*)t->buf;
-	t->storage = PL_CHARS_LOCAL;
-      } else
-      { t->text.w = PL_malloc(len*sizeof(pl_wchar_t));
-	t->storage = PL_CHARS_MALLOC;
-      }
-      succeed;
-    default:
-      assert(0);
-      fail;
-  }
-}
+get_chr_from_text(const PL_chars_t *t, size_t *index)
+{ if ( *index >= t->length )
+    return -1;
 
-
-static inline wint_t
-get_chr_from_text(const PL_chars_t *t, size_t index)
-{ switch(t->encoding)
+  switch(t->encoding)
   { case ENC_ISO_LATIN_1:
-      return t->text.t[index]&0xff;
+      return t->text.t[(*index)++]&0xff;
     case ENC_WCHAR:
-      return t->text.w[index];
+    { int c = t->text.w[(*index)++];
+#if SIZEOF_WCHAR_T == 2
+      if ( IS_UTF16_LEAD(c) )
+      { int c2 = t->text.w[(*index)++];
+	c = utf16_decode(c, c2);
+      }
+#endif
+      return c;
+    }
     default:
       assert(0);
       return 0;
@@ -528,7 +544,37 @@ get_chr_from_text(const PL_chars_t *t, size_t index)
 }
 
 
-#define modify_case_atom(in, out, down, text_type) LDFUNC(modify_case_atom, in, out, down, text_type)
+#define modify_case_atom(in, out, down, text_type) \
+	LDFUNC(modify_case_atom, in, out, down, text_type)
+
+static void
+modify_case_latin_1_to_wide(PL_chars_t *tin, PL_chars_t *tout, Buffer b, int down)
+{ const unsigned char *in = (const unsigned char*)tin->text.t;
+  PL_free_text(tout);
+  initBuffer(b);
+
+  if ( down )
+  { for(size_t i=0; i<tin->length; i++)
+    { wint_t c = towlower(in[i]);
+
+      addWcharBuffer(b, c);
+    }
+  } else				/* upcase */
+  { for(size_t i=0; i<tin->length; i++)
+    { wint_t c = towupper(in[i]);
+
+      addWcharBuffer(b, c);
+    }
+  }
+
+  tout->storage   = PL_CHARS_HEAP;
+  tout->text.w    = baseBuffer(b, wchar_t);
+  tout->length    = entriesBuffer(b, wchar_t);
+  tout->encoding  = ENC_WCHAR;
+  tout->canonical = FALSE;
+}
+
+
 static foreign_t
 modify_case_atom(DECL_LD term_t in, term_t out, int down, int text_type)
 { PL_chars_t tin, tout;
@@ -537,82 +583,100 @@ modify_case_atom(DECL_LD term_t in, term_t out, int down, int text_type)
     return FALSE;
 
   if ( PL_get_text(out, &tout, CVT_ATOMIC) )
-  { unsigned int i;
+  { size_t i1=0, i2=0;
 
-    if ( tin.length != tout.length )
-      fail;
+    for(;;)
+    { int ci = get_chr_from_text(&tin, &i1);
+      int co = get_chr_from_text(&tout, &i2);
 
-    for(i=0; i<tin.length; i++)
-    { wint_t ci = get_chr_from_text(&tin, i);
-      wint_t co = get_chr_from_text(&tout, i);
+      if ( ci == -1 || co == -1 )
+	return ci == co;
 
       if ( down )
-      { if ( co != towlower(ci) )
-	  fail;
+      { if ( co != ftolower(ci) )
+	  return FALSE;
       } else
-      { if ( co != towupper(ci) )
-	  fail;
+      { if ( co != ftoupper(ci) )
+	  return FALSE;
       }
     }
 
-    succeed;
+    return TRUE;
   } else if ( PL_is_variable(out) )
-  { unsigned int i;
-
-    tout.encoding  = tin.encoding;
-    tout.length    = tin.length;
-    tout.canonical = FALSE;		/* or TRUE? Can WCHAR map to ISO? */
-
-    init_tout(&tout, tin.length);
+  { tmp_buffer b;
+    int rc;
 
     if ( tin.encoding == ENC_ISO_LATIN_1 )
     { const unsigned char *in = (const unsigned char*)tin.text.t;
+      size_t i;
+
+      tout.encoding  = tin.encoding;
+      tout.length    = tin.length;
+      tout.canonical = TRUE;
+      if ( tout.length < sizeof(tout.buf) )
+      { tout.text.t  = tout.buf;
+	tout.storage = PL_CHARS_LOCAL;
+      } else
+      { tout.text.t  = PL_malloc(tout.length);
+	tout.storage = PL_CHARS_MALLOC;
+      }
 
       if ( down )
       { for(i=0; i<tin.length; i++)
 	{ wint_t c = towlower(in[i]);
 
-	  if ( c > 255 )
-	  { PL_promote_text(&tout);
-	    for( ; i<tin.length; i++)
-	    { tout.text.w[i] = towlower(in[i]);
-	    }
+	  if ( c > 0xff )
+	  { modify_case_latin_1_to_wide(&tin, &tout, (Buffer)&b, down);
 	    break;
-	  } else
-	  { tout.text.t[i] = (char)c;
 	  }
+	  tout.text.t[i] = (char)c;
 	}
       } else				/* upcase */
       { for(i=0; i<tin.length; i++)
 	{ wint_t c = towupper(in[i]);
 
-	  if ( c > 255 )
-	  { PL_promote_text(&tout);
-	    for( ; i<tin.length; i++)
-	    { tout.text.w[i] = towupper(in[i]);
-	    }
+	  if ( c > 0xff )
+	  { modify_case_latin_1_to_wide(&tin, &tout, (Buffer)&b, down);
 	    break;
-	  } else
-	  { tout.text.t[i] = (char)c;
 	  }
+	  tout.text.t[i] = (char)c;
 	}
       }
-    } else
-    { if ( down )
-      { for(i=0; i<tin.length; i++)
-	{ tout.text.w[i] = towlower(tin.text.w[i]);
+    } else				/* ENC_WCHAR */
+    { const wchar_t* s = (const wchar_t*)tin.text.w;
+      const wchar_t* e = &s[tin.length];
+      int c;
+
+      initBuffer(&b);
+
+      if ( down )
+      { while( s < e )
+	{ s = get_wchar(s, &c);
+	  c = ftolower(c);
+	  addWcharBuffer((Buffer)&b, c);
 	}
       } else
-      { for(i=0; i<tin.length; i++)
-	{ tout.text.w[i] = towupper(tin.text.w[i]);
+      { while( s < e )
+	{ s = get_wchar(s, &c);
+	  c = ftoupper(c);
+	  addWcharBuffer((Buffer)&b, c);
 	}
       }
+
+      tout.storage   = PL_CHARS_HEAP;
+      tout.text.w    = baseBuffer(&b, wchar_t);
+      tout.length    = entriesBuffer(&b, wchar_t);
+      tout.encoding  = ENC_WCHAR;
+      tout.canonical = FALSE;
     }
 
-    PL_unify_text(out, 0, &tout, text_type);
+    rc = PL_unify_text(out, 0, &tout, text_type);
+    PL_free_text(&tin);
     PL_free_text(&tout);
+    if ( tout.encoding != ENC_ISO_LATIN_1 )
+      discardBuffer(&b);
 
-    succeed;
+    return rc;
   } else
   { return PL_error(NULL, 0, NULL, ERR_TYPE,
 		    text_type == PL_STRING ? ATOM_string : ATOM_atom,
@@ -653,44 +717,70 @@ PRED_IMPL("string_upper", 2, string_upper, 0)
 		 *	    WHITE SPACE		*
 		 *******************************/
 
+static void
+skip_separators(PL_chars_t *t, size_t *i)
+{ for(;;)
+  { size_t at = *i;
+    int c  = get_chr_from_text(t, &at);
+
+    if ( c == -1 )
+      return;
+
+    if ( unicode_separator(c) )
+      *i=at;
+    else
+      return;
+  }
+}
+
+
+static int
+copy_non_separators(IOSTREAM *out, PL_chars_t *t, size_t *i)
+{ for(;;)
+  { size_t at = *i;
+    int c  = get_chr_from_text(t, &at);
+
+    if ( c == -1 )
+      return TRUE;
+
+    if ( !unicode_separator(c) )
+    { *i=at;
+      if ( Sputcode(c, out) < 0 )
+	return FALSE;
+    } else
+      return TRUE;
+  }
+}
+
+
 static int
 write_normalize_space(IOSTREAM *out, term_t in)
 { GET_LD
   PL_chars_t tin;
-  size_t i, end;
+  size_t i = 0, end;
 
   if ( !PL_get_text(in, &tin, CVT_ATOMIC|CVT_EXCEPTION) )
     return FALSE;
 
   end = tin.length;
-  i = 0;
-
-  while(i<end && unicode_separator(get_chr_from_text(&tin, i)))
-    i++;
+  skip_separators(&tin, &i);
   while( i<end )
-  { wint_t c;
-
-    while(i<end && !unicode_separator((c=get_chr_from_text(&tin, i))))
-    { if ( Sputcode(c, out) < 0 )
-	fail;
-      i++;
-    }
-    while(i<end && unicode_separator(get_chr_from_text(&tin, i)))
-      i++;
+  { copy_non_separators(out, &tin, &i);
+    skip_separators(&tin, &i);
     if ( i < end )
     { if (  Sputcode(' ', out) < 0 )
-	fail;
+	return FALSE;
     }
   }
 
-  succeed;
+  return TRUE;
 }
 
 
 static
 PRED_IMPL("normalize_space", 2, normalize_space, 0)
 { redir_context ctx;
-  word rc;
+  foreign_t rc;
 
   if ( (rc = setupOutputRedirect(A1, &ctx, FALSE)) )
   { if ( (rc = write_normalize_space(ctx.stream, A2)) )
@@ -711,6 +801,43 @@ PRED_IMPL("normalize_space", 2, normalize_space, 0)
 #if defined(HAVE_LOCALE_H) && defined(HAVE_SETLOCALE)
 #include <locale.h>
 
+typedef struct
+{ const char *name;
+  IOENC encoding;
+} enc_map;
+
+static const enc_map map[] =
+{ { "UTF-8",	  ENC_UTF8 },
+  { "utf8",	  ENC_UTF8 },
+  { "ISO8859-1",  ENC_ISO_LATIN_1 },
+  { "ISO8859_1",  ENC_ISO_LATIN_1 },
+  { "iso88591",   ENC_ISO_LATIN_1 },
+  { "iso_8859_1", ENC_ISO_LATIN_1 },
+  { NULL, ENC_UNKNOWN }
+};
+
+static IOENC
+enc_from_locale(const char *enc)
+{ const char *ext = strchr(enc, '.');
+  const enc_map *m;
+
+  if ( ext )
+    ext++;
+  else
+    ext = enc;
+
+  DEBUG(MSG_LOCALE, Sdprintf("Set encoding from \"%s\"\n", ext));
+
+  for ( m=map; m->name; m++ )
+  { if ( strcasecmp(ext, m->name) == 0 )
+    { return m->encoding;
+    }
+  }
+
+  return ENC_ANSI;			/* text encoding */
+}
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Note: on some installations, locale doesn't   work correctly. Printing a
 message isn't really cute. It would be better to use printMessage(), but
@@ -722,14 +849,27 @@ that locale support is broken. We don't   depend too much on the others,
 so we ignore possible problems.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-static int
+static IOENC
 init_locale(void)
-{ int rc = TRUE;
+{ int rc = ENC_UNKNOWN;
 
-  if ( !setlocale(LC_CTYPE, "") )
-  { rc = FALSE;
-    DEBUG(0, Sdprintf("Failed to set LC_CTYPE locale\n"));
+#ifdef __WINDOWS__
+  UINT cp = GetACP();
+
+  if ( cp == 65001 )
+    rc = ENC_UTF8;
+#endif
+
+  if ( rc == ENC_UNKNOWN && setlocale(LC_CTYPE, "") )
+  { char *enc;
+
+    if ( (enc = setlocale(LC_CTYPE, NULL)) )
+      rc = enc_from_locale(enc);
+  } else if ( rc == ENC_UNKNOWN )
+  { rc = ENC_ISO_LATIN_1;
+    DEBUG(0, Sdprintf("Failed to set LC_CTYPE locale: %s\n", strerror(errno)));
   }
+
   if ( !setlocale(LC_TIME, "") )
   { DEBUG(0, Sdprintf("Failed to set LC_TIME locale\n"));
   }
@@ -863,48 +1003,19 @@ const char _PL_char_types[] = {
 };
 
 
-typedef struct
-{ const char *name;
-  IOENC encoding;
-} enc_map;
-
-static const enc_map map[] =
-{ { "UTF-8",	  ENC_UTF8 },
-  { "utf8",	  ENC_UTF8 },
-  { "ISO8859-1",  ENC_ISO_LATIN_1 },
-  { "ISO8859_1",  ENC_ISO_LATIN_1 },
-  { "iso88591",   ENC_ISO_LATIN_1 },
-  { "iso_8859_1", ENC_ISO_LATIN_1 },
-  { NULL, ENC_UNKNOWN }
-};
-
 IOENC
 initEncoding(void)
 { GET_LD
 
   if ( HAS_LD )
   { if ( !LD->encoding )
-    { char *enc;
+    {
+#if __APPLE__
+      if ( !getenv("LANG") && !getenv("LC_CTYPE") )
+	putenv("LC_CTYPE=UTF-8");
+#endif
 
-      if ( !init_locale() )
-      { LD->encoding = ENC_ISO_LATIN_1;
-      } else if ( (enc = setlocale(LC_CTYPE, NULL)) )
-      { LD->encoding = ENC_ANSI;		/* text encoding */
-
-	if ( (enc = strchr(enc, '.')) )
-	{ const enc_map *m;
-	  enc++;				/* skip '.' */
-
-	  for ( m=map; m->name; m++ )
-	  { if ( strcmp(enc, m->name) == 0 )
-	    { LD->encoding = m->encoding;
-	      break;
-	    }
-	  }
-	}
-      } else
-      { LD->encoding = ENC_ISO_LATIN_1;
-      }
+      LD->encoding = init_locale();
     }
 
     return LD->encoding;
@@ -918,4 +1029,3 @@ void
 initCharTypes(void)
 { initEncoding();
 }
-
