@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2013-2021, VU University Amsterdam
+    Copyright (c)  2013-2022, VU University Amsterdam
 			      SWI-Prolog Solutions b.v.
     All rights reserved.
 
@@ -73,18 +73,14 @@ static int PL_get_dict_ex(term_t data, term_t tag, term_t dict, int flags);
 #define DICT_GET_COLON	0x04
 #define DICT_GET_TERM	0x08
 
-#define CACHED_DICT_FUNCTORS 128
-
-static functor_t dict_functors[CACHED_DICT_FUNCTORS] = {0};
-
 functor_t
 dict_functor(int pairs)
 { if ( pairs < CACHED_DICT_FUNCTORS )
-  { if ( dict_functors[pairs] )
-      return dict_functors[pairs];
+  { if ( GD->dict.dict_functors[pairs] )
+      return GD->dict.dict_functors[pairs];
 
-    dict_functors[pairs] = lookupFunctorDef(ATOM_dict, pairs*2+1);
-    return dict_functors[pairs];
+    GD->dict.dict_functors[pairs] = lookupFunctorDef(ATOM_dict, pairs*2+1);
+    return GD->dict.dict_functors[pairs];
   } else
   { return lookupFunctorDef(ATOM_dict, pairs*2+1);
   }
@@ -763,7 +759,7 @@ PL_get_dict_ex(term_t data, term_t tag, term_t dict, int flags)
       assert(rc == -2);
       Undo(m);
       return ( (ex = PL_new_term_ref()) &&
-	       _PL_unify_atomic(ex, dupl) &&
+	       PL_unify_atomic(ex, dupl) &&
 	       PL_error(NULL, 0, NULL, ERR_DUPLICATE_KEY, ex) );
     }
   }					/* TBD: {name:value, ...} */
@@ -773,10 +769,10 @@ PL_get_dict_ex(term_t data, term_t tag, term_t dict, int flags)
 
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-PL_for_dict()  runs  func  on  each  key-value    pair  in  dict.  Returns
+PL_for_dict()  runs  func  on  each  key-value  pair  in  dict.  Returns
 immediately with the return value of func   if func returns non-zero. If
-the flag DICT_SORTED is given, the  key-value   pairs  are  called in the
-standard order of terms.
+the flag PL_FOR_DICT_SORTED is given, the  key-value pairs are called in
+the standard order of terms.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 typedef struct cmp_dict_index_data
@@ -807,7 +803,7 @@ cmp_dict_index(DECL_LD const int *ip1, const int *ip2, cmp_dict_index_data *ctx)
   } else
   { if ( isAtom(*p) )
     { if ( isAtom(*q) )
-	rc = compareAtoms(*p, *q);
+	rc = compareAtoms(word2atom(*p), word2atom(*q));
       else
 	rc = CMP_GREATER;
     } else
@@ -821,6 +817,32 @@ cmp_dict_index(DECL_LD const int *ip1, const int *ip2, cmp_dict_index_data *ctx)
   return rc;
 }
 
+typedef struct dict_helper
+{ int (*function)(term_t key, term_t value, void *closure);
+  void *closure;
+} dict_helper;
+
+#define public_for_dict_helper(k,v,l,c) \
+	LDFUNC(public_for_dict_helper, k,v,l,c)
+
+static int
+public_for_dict_helper(DECL_LD term_t key, term_t value, int last, void *closure)
+{ struct dict_helper *h = closure;
+
+  return (*h->function)(key, value, h->closure);
+}
+
+int
+PL_for_dict(term_t dict,
+	    int (*func)(term_t key, term_t value, void *closure),
+	    void *closure,
+	    int flags)
+{ GET_LD
+  struct dict_helper h = {.function = func, .closure = closure};
+
+  return _PL_for_dict(dict, public_for_dict_helper, &h, flags);
+}
+
 
 int
 pl_for_dict(DECL_LD term_t dict,
@@ -838,7 +860,7 @@ pl_for_dict(DECL_LD term_t dict,
   arity = arityTerm(*p);
   pairs = arity/2;
 
-  if ( (flags&DICT_SORTED) )
+  if ( (flags&PL_FOR_DICT_SORTED) )
   { cmp_dict_index_data ctx;
 
     if ( pairs < 256 )
@@ -888,9 +910,9 @@ pl_for_dict(DECL_LD term_t dict,
 
 /* resortDictsInClause() resorts the contents of dicts in a clause
 
-This predicate is called from pl-wic.c after   reloading a clause from a
-.qlf file or state if pl-wic.c  detected   a  dict  inside the clause. It
-identifies the code ranges that constitute the  k-v pairs in the dict and
+This predicate is called from pl-qlf.c after   reloading a clause from a
+.qlf file or state if pl-qlf.c  detected   a  dict inside the clause. It
+identifies the code ranges that constitute the k-v pairs in the dict and
 re-orders them according to the new atom-handle ordering.
 
 There    is    a    complicating     factor    with    B_FIRSTVAR/B_VAR,
@@ -990,6 +1012,11 @@ resortDictsInCodes(Code PC, Code end)
 	    return PL_no_memory();
 
 	  PC = stepPC(PC);		/* skip *_FUNCTOR */
+	  if ( fetchop(PC) == H_VOID_N ) /* deal with _{key:_, ...} */
+	  { assert(PC[1] == (code)2);
+	    PC[0] = encode(H_VOID);
+	    PC[1] = encode(H_VOID);
+	  }
 	  PC = skipArgs(PC, 1);		/* skip the type */
 	  fields_start = PC;
 
@@ -1235,7 +1262,7 @@ pl_get_dict(term_t PL__t0, int PL__ac, int ex, control_t PL__ctx)
 	  deRef2(&f->arguments[i+1], np);	/* TBD: check type */
 	  if ( unify_ptrs(&f->arguments[i], valTermRef(A3),
 			  ALLOW_GC|ALLOW_SHIFT) &&
-	       _PL_unify_atomic(A1, *np) )
+	       PL_unify_atomic(A1, *np) )
 	  { PL_close_foreign_frame(fid);
 
 	    if ( i+2 < arity )
@@ -1375,7 +1402,7 @@ PRED_IMPL("dict_pairs", 3, dict_pairs, 0)
 
       if ( PL_get_arg(1, dict, ctx.tmp) &&
 	   PL_unify(ctx.tmp, A2) &&
-	   PL_for_dict(dict, put_pair, &ctx, DICT_SORTED) == 0 )
+	   _PL_for_dict(dict, put_pair, &ctx, PL_FOR_DICT_SORTED) == 0 )
 	return PL_unify_nil_ex(ctx.tail);
 
       return FALSE;
