@@ -3,9 +3,10 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2005-2022, University of Amsterdam
+    Copyright (c)  2005-2024, University of Amsterdam
 			      VU University Amsterdam
 			      CWI, Amsterdam
+			      SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -45,6 +46,7 @@
 #include "pl-gc.h"
 #include "pl-attvar.h"
 #include "pl-inline.h"
+#include "pl-comp.h"
 #undef LD
 #define LD LOCAL_LD
 
@@ -60,7 +62,7 @@ static mpz_t MPZ_MAX_TAGGED;
 static mpz_t MPZ_MIN_PLINT;		/* Prolog int64_t integers */
 static mpz_t MPZ_MAX_PLINT;
 static mpz_t MPZ_MAX_UINT64;
-#if SIZEOF_LONG	< SIZEOF_VOIDP
+#if SIZEOF_LONG	< SIZEOF_WORD
 static mpz_t MPZ_MIN_LONG;		/* Prolog int64_t integers */
 static mpz_t MPZ_MAX_LONG;
 #endif
@@ -86,61 +88,127 @@ static void *(*smp_realloc)(void *, size_t, size_t);
 static void  (*smp_free)(void *, size_t);
 
 #define NOT_IN_PROLOG_ARITHMETIC() \
-	(LD == NULL || LD->gmp.context == NULL || LD->gmp.persistent)
+	(LD == NULL || LD->gmp.context == NULL)
 
-static int
-gmp_too_big(void)
-{ GET_LD
+#define ROUND_SIZE(n) (((n) + (sizeof(size_t) - 1))/sizeof(size_t))
 
-  DEBUG(1, Sdprintf("Signalling GMP overflow\n"));
+typedef enum
+{ GMO_TB_OK = 0,
+  GMP_TB_RESTRAINT,
+  GMP_TB_STACK,
+  GMP_TB_MALLOC
+} gmp_tb;
 
-  return (int)outOfStack((Stack)&LD->stacks.global, STACK_OVERFLOW_THROW);
+#define gmp_too_big(why)      LDFUNC(gmp_too_big, why)
+#define gmp_check_size(bytes) LDFUNC(gmp_check_size, bytes)
+
+static void*			/* Actually, does not return */
+gmp_too_big(DECL_LD gmp_tb why)
+{ DEBUG(MSG_GMP_OVERFLOW, Sdprintf("Signalling GMP overflow\n"));
+
+  switch(why)
+  { case GMP_TB_RESTRAINT:
+    { number max = { .type = V_INTEGER };
+      max.value.i = LD->gmp.max_integer_size;
+      PL_error(NULL, 0, "requires more than max_integer_size bytes",
+	       ERR_AR_TRIPWIRE, ATOM_max_integer_size, &max);
+      PL_rethrow();
+    }
+    case GMP_TB_STACK:
+      outOfStack((Stack)&LD->stacks.global, STACK_OVERFLOW_THROW);
+      break;
+    case GMP_TB_MALLOC:
+    default:
+      PL_no_memory();
+      PL_rethrow();
+  }
+
+  abortProlog();		/* Just in case the above fails */
+  PL_rethrow();
+  return FALSE;
 }
 
-#define TOO_BIG_GMP(n) ((n) > 1000 && (n) > (size_t)globalStackLimit())
+
+static int
+gmp_check_size(DECL_LD size_t bytes)
+{ gmp_tb why = GMO_TB_OK;
+
+  if ( bytes <= 1000 )
+    return TRUE;
+  if ( bytes > LD->gmp.max_integer_size )
+    why = GMP_TB_RESTRAINT;
+  else if ( bytes > (size_t)globalStackLimit())
+    why = GMP_TB_STACK;
+  else
+    return TRUE;
+
+  gmp_too_big(why);
+  return FALSE;
+}
+
+
+#define TOO_BIG_GMP(n) ((n) > 1000 &&			   \
+			((n) > LD->gmp.max_integer_size || \
+			 (n) > (size_t)globalStackLimit()))
 
 static void *
 mp_alloc(size_t bytes)
 { GET_LD
+  ar_context *ctx;
   mp_mem_header *mem;
 
   if ( NOT_IN_PROLOG_ARITHMETIC() )
     return smp_alloc(bytes);
 
-  if ( TOO_BIG_GMP(bytes) ||
-       !(mem = malloc(sizeof(mp_mem_header)+bytes)) )
-  { gmp_too_big();
-    abortProlog();
-    PL_rethrow();
-    return NULL;			/* make compiler happy */
-  }
+  if ( !gmp_check_size(bytes) )
+    return NULL;
 
 #if O_BF
   if ( bytes == 0 )
     return NULL;
 #endif
+  ctx = LD->gmp.context;
 
-  GMP_LEAK_CHECK(LD->gmp.allocated += bytes);
-
-  mem->next = NULL;
-  mem->context = LD->gmp.context;
-  if ( LD->gmp.tail )
-  { mem->prev = LD->gmp.tail;
-    LD->gmp.tail->next = mem;
-    LD->gmp.tail = mem;
-  } else
-  { mem->prev = NULL;
-    LD->gmp.head = LD->gmp.tail = mem;
+  size_t fastunits = ROUND_SIZE(bytes)+1;
+  if ( ctx->allocated+fastunits <= GMP_STACK_ALLOC )
+  { size_t *data = &ctx->alloc_buf[ctx->allocated];
+    *data++ = fastunits;
+    ctx->allocated += fastunits;
+    DEBUG(MSG_GMP_ALLOC, Sdprintf("GMP: from stack %zd@%p\n", bytes, data));
+    return data;
   }
-  DEBUG(9, Sdprintf("GMP: alloc %ld@%p\n", bytes, &mem[1]));
 
-  return &mem[1];
+  if ( (mem = tmp_malloc(sizeof(mp_mem_header)+bytes)) )
+  { mem->next = NULL;
+    if ( ctx->tail )
+    { mem->prev = ctx->tail;
+      ctx->tail->next = mem;
+      ctx->tail = mem;
+    } else
+    { mem->prev = NULL;
+      ctx->head = ctx->tail = mem;
+    }
+    DEBUG(MSG_GMP_ALLOC, Sdprintf("GMP: malloc %zd@%p\n", bytes, &mem[1]));
+
+    return &mem[1];
+  } else
+    return gmp_too_big(GMP_TB_MALLOC);
+}
+
+
+static inline size_t *
+mp_on_stack(ar_context *ctx, void *ptr)
+{ if ( ptr > (void*)ctx->alloc_buf && ptr < (void*)&ctx->alloc_buf[GMP_STACK_ALLOC] )
+    return &((size_t*)ptr)[-1];
+
+  return NULL;
 }
 
 
 static void
 mp_free(void *ptr, size_t size)
 { GET_LD
+  ar_context *ctx;
   mp_mem_header *mem;
 
   if ( NOT_IN_PROLOG_ARITHMETIC() )
@@ -152,32 +220,39 @@ mp_free(void *ptr, size_t size)
   if ( !ptr )
     return;
 #endif
+  ctx = LD->gmp.context;
+  size_t *base = mp_on_stack(ctx, ptr);
+  if ( base )
+  { if ( (base-ctx->alloc_buf) + base[0] == ctx->allocated )
+      ctx->allocated -= base[0];
+    return;
+  }
 
   mem = ((mp_mem_header*)ptr)-1;
 
-  if ( mem == LD->gmp.head )
-  { LD->gmp.head = LD->gmp.head->next;
-    if ( LD->gmp.head )
-      LD->gmp.head->prev = NULL;
+  if ( mem == ctx->head )
+  { ctx->head = ctx->head->next;
+    if ( ctx->head )
+      ctx->head->prev = NULL;
     else
-      LD->gmp.tail = NULL;
-  } else if ( mem == LD->gmp.tail )
-  { LD->gmp.tail = LD->gmp.tail->prev;
-    LD->gmp.tail->next = NULL;
+      ctx->tail = NULL;
+  } else if ( mem == ctx->tail )
+  { ctx->tail = ctx->tail->prev;
+    ctx->tail->next = NULL;
   } else
   { mem->prev->next = mem->next;
     mem->next->prev = mem->prev;
   }
 
-  free(mem);
-  DEBUG(9, Sdprintf("GMP: free: %ld@%p\n", size, ptr));
-  GMP_LEAK_CHECK(LD->gmp.allocated -= size);
+  tmp_free(mem);
+  DEBUG(MSG_GMP_ALLOC, Sdprintf("GMP: free: %zd@%p\n", size, ptr));
 }
 
 
 static void *
 mp_realloc(void *ptr, size_t oldsize, size_t newsize)
 { GET_LD
+  ar_context *ctx;
   mp_mem_header *oldmem, *newmem;
 
   if ( NOT_IN_PROLOG_ARITHMETIC() )
@@ -192,54 +267,107 @@ mp_realloc(void *ptr, size_t oldsize, size_t newsize)
   }
 #endif
 
+  if ( newsize > oldsize && !gmp_check_size(newsize) )
+    return NULL;
+
+  ctx = LD->gmp.context;
+  size_t *base = mp_on_stack(ctx, ptr);
+  if ( base )
+  { size_t fastunits = ROUND_SIZE(newsize)+1;
+    size_t alloc0 = base-ctx->alloc_buf;
+
+    if ( alloc0 + base[0] == ctx->allocated ) /* at the top */
+    { if ( alloc0+fastunits-1 <= GMP_STACK_ALLOC )
+      { base[0] = fastunits;		      /* and still fits */
+	ctx->allocated = alloc0+fastunits;
+	return ptr;
+      }
+    } else if ( fastunits <= base[0] )	      /* shrink */
+    { base[0] = fastunits;
+      return ptr;
+    }
+
+    void *new = mp_alloc(newsize);
+    if ( new )
+    { size_t cp = base[0]*sizeof(size_t);
+      if ( newsize < cp )
+	cp = newsize;
+      memcpy(new, ptr, cp);
+    }
+    return new;
+  }
+
   oldmem = ((mp_mem_header*)ptr)-1;
-  if ( TOO_BIG_GMP(newsize) ||
-       !(newmem = realloc(oldmem, sizeof(mp_mem_header)+newsize)) )
-  { gmp_too_big();
-    abortProlog();
-    PL_rethrow();
-    return NULL;			/* make compiler happy */
-  }
+  if ( (newmem = tmp_realloc(oldmem, sizeof(mp_mem_header)+newsize)) )
+  { if ( oldmem != newmem )		/* re-link if moved */
+    { if ( newmem->prev )
+	newmem->prev->next = newmem;
+      else
+	ctx->head = newmem;
 
-  if ( oldmem != newmem )		/* re-link if moved */
-  { if ( newmem->prev )
-      newmem->prev->next = newmem;
-    else
-      LD->gmp.head = newmem;
+      if ( newmem->next )
+	newmem->next->prev = newmem;
+      else
+	ctx->tail = newmem;
+    }
 
-    if ( newmem->next )
-      newmem->next->prev = newmem;
-    else
-      LD->gmp.tail = newmem;
-  }
-
-  GMP_LEAK_CHECK(LD->gmp.allocated -= oldsize;
-		 LD->gmp.allocated += newsize);
-  DEBUG(9, Sdprintf("GMP: realloc %ld@%p --> %ld@%p\n", oldsize, ptr, newsize, &newmem[1]));
-
-  return &newmem[1];
+    DEBUG(MSG_GMP_ALLOC, Sdprintf("GMP: realloc %zd@%p --> %zd@%p\n", oldsize, ptr, newsize, &newmem[1]));
+    return &newmem[1];
+  } else
+    return gmp_too_big(GMP_TB_MALLOC);
 }
 
 
 void
 mp_cleanup(ar_context *ctx)
-{ GET_LD
-  mp_mem_header *mem, *next;
+{ mp_mem_header *mem, *next;
 
-  if ( LD->gmp.context )
-  { for(mem=LD->gmp.head; mem; mem=next)
-    { next = mem->next;
-      if ( mem->context == LD->gmp.context )
-      { DEBUG(9, Sdprintf("GMP: cleanup of %p\n", &mem[1]));
-	mp_free(&mem[1], 0);
-      }
-    }
+  for(mem=ctx->head; mem; mem=next)
+  { next = mem->next;
+    DEBUG(MSG_GMP_ALLOC, Sdprintf("GMP: cleanup of %p\n", &mem[1]));
+    mp_free(&mem[1], 0);
   }
+}
 
-  LD->gmp.context = ctx->parent;
+#ifdef O_DEBUG
+static int
+mp_test_alloc(void)
+{ GET_LD
+  AR_CTX;
+
+  AR_BEGIN();
+
+  char *first  = mp_alloc(8);
+  strcpy(first, "hello");
+  char *second = mp_alloc(8);
+  assert(second-first == sizeof(size_t)+8);
+  /* realloc top */
+  char *ext = mp_realloc(second, 8, 12);
+  assert(ext == second);
+  assert(__PL_ar_ctx.allocated == ROUND_SIZE(8)+1+ROUND_SIZE(12)+1);
+  /* free top */
+  mp_free(ext, 12);
+  assert(__PL_ar_ctx.allocated == ROUND_SIZE(8)+1);
+  /* re-add second */
+  second = mp_alloc(8);
+  assert(second-first == sizeof(size_t)+8);
+  /* realloc non-first (move) */
+  ext = mp_realloc(first, 8, 25);
+  assert(ext-second == sizeof(size_t)+8);
+  assert(strcmp(ext, "hello") == 0);
+  /* shrink last */
+  char *ext2 = mp_realloc(ext, 25, 8);
+  assert(ext == ext2);
+  assert(mp_on_stack(&__PL_ar_ctx, ext2)[0] == ROUND_SIZE(8)+1);
+
+  AR_END();
+
+  return TRUE;
 }
 #endif
 
+#endif /*O_MY_GMP_ALLOC*/
+
 
 #ifdef __WINDOWS__
 #undef isascii
@@ -462,159 +590,128 @@ the global stack.
 Normally called through the inline get_integer() function.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-void
-get_bigint(word w, Number n)
-{ GET_LD
-  Word p = addressIndirect(w);
-  size_t wsize = wsizeofInd(*p);
-
-  DEBUG(0, assert(storage(w) == STG_INLINE));
-
-  p++;
-  if ( wsize == WORDS_PER_INT64 )
-  { n->type = V_INTEGER;
-    memcpy(&n->value.i, p, sizeof(int64_t));
-  } else
-  { n->type = V_MPZ;
-
+static void
+get_mpz_from_stack(Word p, mpz_t mpz)
+{
 #if O_GMP
-    n->value.mpz->_mp_size  = mpz_stack_size(*p++);
-    n->value.mpz->_mp_alloc = 0;
-    n->value.mpz->_mp_d     = (mp_limb_t*) p;
+  mpz->_mp_size  = mpz_stack_size(*p++);
+  mpz->_mp_alloc = 0;
+  mpz->_mp_d     = (mp_limb_t*) p;
 #elif O_BF
-    slimb_t len = mpz_stack_size(*p++);
-    n->value.mpz->ctx	 = NULL;
-    n->value.mpz->expn = (slimb_t)*p++;
-    n->value.mpz->sign = len < 0;
-    n->value.mpz->len  = abs(len);
-    n->value.mpz->tab  = (limb_t*)p;
+  slimb_t len = mpz_stack_size(*p++);
+  mpz->ctx	 = NULL;
+  mpz->expn = (slimb_t)*p++;
+  mpz->sign = len < 0;
+  mpz->len  = abs(len);
+  mpz->tab  = (limb_t*)p;
 #endif
-  }
 }
 
+static void
+get_mpq_from_stack(Word p, mpq_t mpq)
+{ mpz_t num, den;
+  size_t num_size;
+
+#if O_GMP
+  num->_mp_size  = mpz_stack_size(*p++);
+  num->_mp_alloc = 0;
+  num->_mp_d     = (mp_limb_t*) (p+1);
+  num_size       = mpz_wsize(num, NULL);
+  den->_mp_size  = mpz_stack_size(*p++);
+  den->_mp_alloc = 0;
+  den->_mp_d     = (mp_limb_t*) (p+num_size);
+#elif O_BF
+  slimb_t len = mpz_stack_size(*p++);
+  num->ctx    = NULL;
+  num->alloc  = 0;
+  num->expn   = (slimb_t)*p++;
+  num->sign   = len < 0;
+  num->len    = abs(len);
+  num->tab    = (limb_t*)(p+2);
+  num_size    = mpz_wsize(num, NULL);
+  den->ctx    = NULL;
+  den->alloc  = 0;
+  den->sign   = 0;			/* canonical MPQ */
+  den->len    = mpz_stack_size(*p++);
+  den->expn   = (slimb_t)*p++;
+  den->tab    = (limb_t*) (p+num_size);
+#endif
+  *mpq_numref(mpq) = num[0];
+  *mpq_denref(mpq) = den[0];
+}
+
+void
+get_bigint(word w, Number n)
+{ Word p = addressIndirect(w);
+
+  DEBUG(0, assert(storage(w) != STG_INLINE));
+
+  p++;				/* bits of indirect */
+  n->type = V_MPZ;
+  get_mpz_from_stack(p, n->value.mpz);
+}
+
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Get a rational (int or non-int-rational) from a Prolog word, knowing the
+word is not an inlined (tagged) integer.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 void
 get_rational_no_int(DECL_LD word w, Number n)
 { Word p = addressIndirect(w);
-  size_t wsize = wsizeofInd(*p);
 
   p++;
-  if ( wsize == WORDS_PER_INT64 )
-  { n->type = V_INTEGER;
-    memcpy(&n->value.i, p, sizeof(int64_t));
-  } else if ( (*p&MP_RAT_MASK) )
-  { mpz_t num, den;
-    size_t num_size;
-
-    n->type = V_MPQ;
-#if O_GMP
-    num->_mp_size  = mpz_stack_size(*p++);
-    num->_mp_alloc = 0;
-    num->_mp_d     = (mp_limb_t*) (p+1);
-    num_size       = mpz_wsize(num, NULL);
-    den->_mp_size  = mpz_stack_size(*p++);
-    den->_mp_alloc = 0;
-    den->_mp_d     = (mp_limb_t*) (p+num_size);
-#elif O_BF
-    slimb_t len = mpz_stack_size(*p++);
-    num->ctx	 = NULL;
-    num->alloc = 0;
-    num->expn  = (slimb_t)*p++;
-    num->sign  = len < 0;
-    num->len   = abs(len);
-    num->tab   = (limb_t*)(p+2);
-    num_size   = mpz_wsize(num, NULL);
-    den->ctx   = NULL;
-    den->alloc = 0;
-    den->sign  = 0;			/* canonical MPQ */
-    den->len   = mpz_stack_size(*p++);
-    den->expn  = (slimb_t)*p++;
-    den->tab   = (limb_t*) (p+num_size);
-#endif
-    *mpq_numref(n->value.mpq) = num[0];
-    *mpq_denref(n->value.mpq) = den[0];
+  if ( (*p&MP_RAT_MASK) )
+  { n->type = V_MPQ;
+    get_mpq_from_stack(p, n->value.mpq);
   } else
   { n->type = V_MPZ;
-
-#if O_GMP
-    n->value.mpz->_mp_size  = mpz_stack_size(*p++);
-    n->value.mpz->_mp_alloc = 0;
-    n->value.mpz->_mp_d     = (mp_limb_t*) p;
-#elif O_BF
-    slimb_t len = mpz_stack_size(*p++);
-    n->value.mpz->ctx	 = NULL;
-    n->value.mpz->expn = (slimb_t)*p++;
-    n->value.mpz->sign = len < 0;
-    n->value.mpz->len  = abs(len);
-    n->value.mpz->tab  = (limb_t*)p;
-#endif
+    get_mpz_from_stack(p, n->value.mpz);
   }
 }
 
 
 Code
 get_mpz_from_code(Code pc, mpz_t mpz)
-{ size_t wsize = wsizeofInd(*pc);
+{ word m;
+  Word data;
+  pc = code_get_indirect(pc, &m, &data);
 
-  pc++;
-#if O_GMP
-  mpz->_mp_size  = mpz_stack_size(*pc);
-  mpz->_mp_alloc = 0;
-  mpz->_mp_d     = (mp_limb_t*)(pc+1);
-#elif O_BF
-  slimb_t len = mpz_stack_size(*pc);
-  mpz->ctx   = NULL;
-  mpz->alloc = 0;
-  mpz->expn  = (slimb_t)pc[1];
-  mpz->sign  = len < 0;
-  mpz->len   = abs(len);
-  mpz->tab   = (limb_t*)pc+2;
-#endif
+  get_mpz_from_stack(data, mpz);
 
-  return pc+wsize;
+  return pc;
 }
 
+
 Code
 get_mpq_from_code(Code pc, mpq_t mpq)
-{ Word p = pc;
-  size_t wsize = wsizeofInd(*p);
-  p++;
-  mpz_t num, den;
+{ word m;
+  Word data;
+  pc = code_get_indirect(pc, &m, &data);
 
-#if O_GMP
-  int num_size = mpz_stack_size(*p++);
-  int den_size = mpz_stack_size(*p++);
-  size_t limpsize = sizeof(mp_limb_t) * abs(num_size);
-  num->_mp_size   = num_size;
-  den->_mp_size   = den_size;
-  num->_mp_alloc  = 0;
-  den->_mp_alloc  = 0;
-  num->_mp_d = (mp_limb_t*)p;
-  p += (limpsize+sizeof(word)-1)/sizeof(word);
-  den->_mp_d = (mp_limb_t*)p;
-#elif O_BF
-  int num_size = mpz_stack_size(*p++);
-  num->ctx = NULL;
-  num->expn = *p++;
-  num->sign = num_size < 0;
-  num->len  = abs(num_size);
-  int den_size = mpz_stack_size(*p++);
-  size_t limpsize = sizeof(mp_limb_t) * abs(num_size);
-  den->ctx = NULL;
-  den->expn = *p++;
-  den->sign = den_size < 0;
-  den->len  = abs(den_size);
-  num->tab = (mp_limb_t*)p;
-  p += (limpsize+sizeof(word)-1)/sizeof(word);
-  den->tab = (mp_limb_t*)p;
-#endif
-  *mpq_numref(mpq) = num[0];
-  *mpq_denref(mpq) = den[0];
+  get_mpq_from_stack(data, mpq);
 
-  return pc+wsize+1;
+  return pc;
 }
 
 
+int
+get_int64(DECL_LD word w, int64_t *ip)
+{ if ( tagex(w) == (TAG_INTEGER|STG_INLINE) )
+  { *ip = valInt(w);
+    return TRUE;
+  } else if ( tagex(w) == (TAG_INTEGER|STG_GLOBAL) )
+  { number n;
+
+    get_bigint(w, &n);
+    return mpz_to_int64(n.value.mpz, ip);
+  } else
+    return FALSE;
+}
+
+
+
 		 /*******************************
 		 *	  IMPORT/EXPORT		*
 		 *******************************/
@@ -754,8 +851,7 @@ by N bytes in big endian notation.
 
 char *
 loadMPZFromCharp(const char *data, Word r, Word *store)
-{ GET_LD
-  int size = 0;
+{ int size = 0;
   size_t limbsize;
   size_t wsize;
   int neg;
@@ -789,8 +885,7 @@ loadMPZFromCharp(const char *data, Word r, Word *store)
 
 char *
 loadMPQFromCharp(const char *data, Word r, Word *store)
-{ GET_LD
-  int num_size;
+{ int num_size;
   int den_size;
   size_t num_limbsize, num_wsize;
   size_t den_limbsize, den_wsize;
@@ -1082,7 +1177,7 @@ initGMP(void)
     mpz_init_set_si64(MPZ_MIN_PLINT, PLMININT);
     mpz_init_set_si64(MPZ_MAX_PLINT, PLMAXINT);
     mpz_init_max_uint(MPZ_MAX_UINT64, 64);
-#if SIZEOF_LONG < SIZEOF_VOIDP
+#if SIZEOF_LONG < SIZEOF_WORD
     mpz_init_set_si64(MPZ_MIN_LONG, LONG_MIN);
     mpz_init_set_si64(MPZ_MAX_LONG, LONG_MAX);
 #endif
@@ -1091,6 +1186,7 @@ initGMP(void)
     { mp_get_memory_functions(&smp_alloc, &smp_realloc, &smp_free);
       mp_set_memory_functions(mp_alloc, mp_realloc, mp_free);
     }
+    DEBUG(0, mp_test_alloc());
 #endif
 
 #if O_GMP
@@ -1133,6 +1229,8 @@ cleanupGMP(void)
    as used  for clause indexing.   If the integer  is huge, we  do not
    want to use the whole thing.   Instead, we pick the dimensions, the
    first two and last limb of the content.
+
+   Note: p might be aligned at Code rather than Word.
  */
 
 word
@@ -1264,15 +1362,13 @@ without any knowledge of the represented data.
 #define put_mpz(at, mpz, flags) LDFUNC(put_mpz, at, mpz, flags)
 static int
 put_mpz(DECL_LD Word at, mpz_t mpz, int flags)
-{ int64_t v;
-
-  DEBUG(2,
+{ DEBUG(2,
 	{ char buf[256];
 	  Sdprintf("put_mpz(%s)\n",
 		   mpz_get_str(buf, 10, mpz));
 	});
 
-#if SIZEOF_LONG < SIZEOF_VOIDP
+#if SIZEOF_LONG < SIZEOF_WORD
   if ( mpz_cmp(mpz, MPZ_MIN_LONG) >= 0 &&
        mpz_cmp(mpz, MPZ_MAX_LONG) <= 0 )
 #else
@@ -1289,9 +1385,8 @@ put_mpz(DECL_LD Word at, mpz_t mpz, int flags)
     }
 
     *at = consInt(v);
+    assert(valInt(*at) == v);
     return TRUE;
-  } else if ( mpz_to_int64(mpz, &v) )
-  { return put_int64(at, v, flags);
   } else
   { return globalMPZ(at, mpz, flags);
   }
@@ -1307,6 +1402,28 @@ put_mpz(DECL_LD Word at, mpz_t mpz, int flags)
   LOCAL_OVERFLOW: cannot represent (no GMP)
 */
 
+int
+put_int64(DECL_LD Word at, int64_t l, int flags)
+{ word r;
+
+  r = consInt(l);
+  if ( valInt(r) == l )
+  { *at = r;
+    return TRUE;
+  } else
+  {
+#ifdef O_BIGNUM
+    mpz_t mpz;
+
+    mpz_init_set_si64(mpz, l);
+    return globalMPZ(at, mpz, flags);
+#else
+    return LOCAL_OVERFLOW;
+#endif
+  }
+}
+
+
 int
 put_uint64(DECL_LD Word at, uint64_t l, int flags)
 { if ( (int64_t)l >= 0 )
@@ -1317,7 +1434,7 @@ put_uint64(DECL_LD Word at, uint64_t l, int flags)
     mpz_t mpz;
 
     mpz_init_set_uint64(mpz, l);
-    return put_mpz(at, mpz, flags);
+    return globalMPZ(at, mpz, flags);
 #else
     return LOCAL_OVERFLOW;
 #endif
@@ -1736,14 +1853,46 @@ cmp_q_q(mpq_t q1, mpq_t q2)
 
 static int
 cmp_z_i(mpz_t z1, int64_t i2)
-{ int t = mpz_cmp_si(z1,i2);
+{
+#if SIZEOF_LONG == 8
+  int t = mpz_cmp_si(z1,i2);
   return (t < 0) ? CMP_LESS : (t > 0);
+#else
+  if ( i2 >= LONG_MIN && i2 <= LONG_MAX )
+  { int t = mpz_cmp_si(z1,(long)i2);
+    return (t < 0) ? CMP_LESS : (t > 0);
+  } else
+  { mpz_t a;
+    mpz_init_set_si64(a, i2);
+    int t = cmp_z_z(z1, a);
+    mpz_clear(a);
+    return t;
+  }
+#endif
 }
 
 static int
 cmp_q_i(mpq_t q1, int64_t i2)
-{ int t = mpq_cmp_si(q1,i2,1);
+{
+#if SIZEOF_LONG == 8
+  int t = mpq_cmp_si(q1,i2,1);
   return (t < 0) ? CMP_LESS : (t > 0);
+#else
+  if ( i2 >= LONG_MIN && i2 <= LONG_MAX )
+  { int t = mpq_cmp_si(q1,(long)i2,1);
+    return (t < 0) ? CMP_LESS : (t > 0);
+  } else
+  { mpq_t qa;
+    mpz_t za;
+    mpq_init(qa);
+    mpz_init_set_si64(za, i2);
+    mpq_set_z(qa, za);
+    int t = cmp_q_q(q1, qa);
+    mpz_clear(za);
+    mpq_clear(qa);
+    return t;
+  }
+#endif
 }
 
 static int
@@ -2123,8 +2272,10 @@ PL_get_mpq(term_t t, mpq_t mpq)
 { if ( PL_is_rational(t) )
   { GET_LD
     number n;
+    Word p = valTermRef(t);
 
-    get_rational(t, &n);
+    deRef(p);
+    get_rational(*p, &n);
     switch(n.type)
     { case V_INTEGER:
 	if ( n.value.i >= LONG_MIN && n.value.i <= LONG_MAX )
