@@ -290,7 +290,9 @@ findHome(const char *symbols, int argc, const char **argv)
     Ssnprintf(buf, sizeof(buf), "%s/" PLHOMEFILE, pparent);
 
     if ( (fd = Sopen_file(buf, "r")) )
-    { if ( Sfgets(buf, sizeof(buf), fd) )
+    { DEBUG(MSG_INITIALISE, Sdprintf("Found home link file %s\n", buf));
+
+      if ( Sfgets(buf, sizeof(buf), fd) )
       { size_t l = strlen(buf);
 
 	while(l > 0 && buf[l-1] <= ' ')
@@ -318,7 +320,7 @@ findHome(const char *symbols, int argc, const char **argv)
 	if ( ExistsDirectory(maybe_home) )
 	{ home = maybe_home;
 	  DEBUG(MSG_INITIALISE,
-		Sdprintf("Found home using %s from %s\n", buf));
+		Sdprintf("Found home %s using link file\n", home));
 	}
       }
       Sclose(fd);
@@ -352,7 +354,7 @@ findHome(const char *symbols, int argc, const char **argv)
 	 ExistsDirectory(maybe_home)
        ) )
   { home = maybe_home;
-    DEBUG(MSG_INITIALISE, Sdprintf("Found home using %s\n", PLHOME));
+    DEBUG(MSG_INITIALISE, Sdprintf("Using compiled-in home at %s\n", PLHOME));
   }
 #endif
 
@@ -527,8 +529,6 @@ initPaths(int argc, const char **argv)
 
     DEBUG(MSG_INITIALISE, Sdprintf("rc-module: %s\n", symbols));
 
-    systemDefaults.home	       = findHome(symbols, argc, argv);
-
 #ifdef __WINDOWS__			/* we want no module but the .EXE */
     GD->paths.module	       = store_string(symbols);
     symbols = findExecutable(NULL, plp, sizeof(plp));
@@ -536,6 +536,8 @@ initPaths(int argc, const char **argv)
 #endif
     GD->paths.executable       = store_string(symbols);
     GD->options.systemInitFile = defaultSystemInitFile(argv[0]);
+
+    systemDefaults.home	       = findHome(symbols, argc, argv);
   } else
   { systemDefaults.home	       = findHome(NULL, argc, argv);
     GD->options.systemInitFile = store_string("none");
@@ -679,7 +681,7 @@ again:
 	discardBuffer(name);
 	goto again;
       default:
-	addBuffer(name, c, char);
+	addBuffer(name, (char)c, char);
     }
   }
 
@@ -691,7 +693,7 @@ do_value:
 	addBuffer(value, EOS, char);
 	return TRUE;
       default:
-	addBuffer(value, c, char);
+	addBuffer(value, (char)c, char);
     }
   }
 }
@@ -1065,7 +1067,7 @@ PL_winitialise(int argc, wchar_t **wargv)
 
     for(w = wargv[i]; *w; w++)
     { if ( *w < 0x7f )
-      { char c = *w;
+      { char c = (char)*w;
 	addBuffer(&b, c, char);
       } else
       { char wb[6];
@@ -1164,6 +1166,14 @@ PL_initialise(int argc, char **argv)
 #endif
   }
   initDefaultOptions();
+  if ( GD->resources.DB )
+  { IOSTREAM *opts;
+
+    if ( (opts=SopenZIP(GD->resources.DB, "$prolog/options.txt", RC_RDONLY)) )
+    { loadStateOptions(opts);
+      Sclose(opts);
+    }
+  }
 
   if ( !GD->resources.DB ||
        !streq(GD->options.saveclass, "runtime") )
@@ -1217,7 +1227,6 @@ PL_initialise(int argc, char **argv)
 	argv = argvleft;
       }
     }
-
   } else
   { argc--;				/* saved state: only drop program */
     argv++;
@@ -1434,6 +1443,7 @@ abi_version(void)
 	  PL_QLF_LOADVERSION,
 	  GD->foreign.signature,
 	  VM_SIGNATURE);
+  PL_cleanup(0);
 
   return TRUE;
 }
@@ -1834,6 +1844,14 @@ action:
   return FALSE;					/* not reached */
 }
 
+void
+PL_system_error(const char *fm, ...)
+{ va_list args;
+
+  va_start(args, fm);
+  vsysError(fm, args);
+  va_end(args);
+}
 
 void
 vfatalError(const char *fm, va_list args)
