@@ -3,8 +3,9 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2012-2019, University of Amsterdam
+    Copyright (c)  2012-2024, University of Amsterdam
 			      CWI, Amsterdam
+			      SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -33,6 +34,11 @@
     POSSIBILITY OF SUCH DAMAGE.
 */
 
+#include "pl-incl.h"
+
+#ifndef _PL_INIT_H
+#define _PL_INIT_H
+
 		/********************************
 		*           STRUCTURES		*
 		********************************/
@@ -70,9 +76,12 @@ typedef struct
   char *	config;			/* Show config info */
   opt_list     *scriptFiles;
   opt_list     *search_paths;		/* -p path */
+  opt_list     *defines;		/* -Dname=value */
   char *	pldoc_server;		/* --pldoc=Server */
+  char *	bootFrom;		/* -x file */
   char *	compileOut;		/* file to store compiler output */
   char *	saveclass;		/* Type of saved state */
+  bool		compile;		/* -c source ... */
   bool		silent;			/* -q: quiet operation */
   bool		traditional;		/* --traditional: no version 7 exts */
   bool		nothreads;		/* --no-threads */
@@ -85,16 +94,25 @@ typedef struct
 #endif
 } pl_options_t;
 
-COMMON(int)	opt_append(opt_list **l, const char *s);
+		 /*******************************
+		 *    FUNCTION DECLARATIONS	*
+		 *******************************/
+
+int		startProlog(int argc, char **argv);
+bool		sysError(const char *fm, ...);
+void		fatalError(const char *fm, ...) NORETURN;
+bool		warning(const char *fm, ...);
+void		vfatalError(const char *fm, va_list args) NORETURN;
+bool		vwarning(const char *fm, va_list args);
+int		run_on_halt(OnHalt *handlers, int rval);
+int		setTraditional(void);
+int		opt_append(opt_list **l, const char *s);
 
 
 		/********************************
 		*           PARAMETERS		*
 		********************************/
 
-#ifndef DEFSTARTUP
-#define DEFSTARTUP "init.pl"
-#endif
 #ifndef SYSTEMHOME
 #define SYSTEMHOME "/usr/lib/swipl"
 #endif
@@ -110,8 +128,8 @@ COMMON(int)	opt_append(opt_list **l, const char *s);
 #endif
 
 
-#define DEF_DEFDEFSTACKLIMIT	(((size_t)1024/8)*1024*1024*SIZEOF_VOIDP)
-#define DEF_DEFTABLE		(((size_t)1024/8)*1024*1024*SIZEOF_VOIDP)
+#define DEF_DEFDEFSTACKLIMIT	(((size_t)1024/8)*1024*1024*SIZEOF_WORD)
+#define DEF_DEFTABLE		(((size_t)1024/8)*1024*1024*SIZEOF_WORD)
 
 #ifndef DEFSTACKLIMIT
 #define DEFSTACKLIMIT   DEF_DEFDEFSTACKLIMIT
@@ -125,3 +143,5 @@ COMMON(int)	opt_append(opt_list **l, const char *s);
 #define PLHOMEVAR_1	"SWI_HOME_DIR"
 #define PLHOMEVAR_2	"SWIPL"
 #define PLHOMEFILE	"swipl.home"
+
+#endif /*_PL_INIT_H*/
