@@ -79,7 +79,6 @@ resetProlog(int clear_stacks)
 
     gc_status.blocked        = 0;
     LD->shift_status.blocked = 0;
-    LD->in_arithmetic        = 0;
     LD->in_print_message     = 0;
   }
 
@@ -154,7 +153,10 @@ query_loop(atom_t goal, int loop)
     }
 
     if ( !rc && (except = PL_exception(qid)) )
-    { if ( Sferror(Suser_input) ||
+    { if ( classify_exception(except) == EXCEPT_ABORT )
+	Sclearerr(Suser_input);
+
+      if ( Sferror(Suser_input) ||
 	   Sferror(Suser_output) ||
 	   Sferror(Suser_error) )
 	return -1;
@@ -190,10 +192,10 @@ the debugger.  Restores I/O and debugger on exit.  The Prolog  predicate
 `$break' is called to actually built the break environment.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-static int
+static foreign_t
 pl_break1(atom_t goal)
 { GET_LD
-  int rc = TRUE;
+  foreign_t rc = TRUE;
   int old_level = LD->break_level;
 
   IOSTREAM *inSave       = Scurin;
@@ -248,13 +250,13 @@ Run a nested toplevel. Do not  use   PRED_IMPL()  for this because it is
 very handy to use from e.g., the gdb debugger.
 */
 
-word
+foreign_t
 pl_break(void)
 { GET_LD
   wakeup_state wstate;
 
   if ( saveWakeup(&wstate, TRUE) )
-  { word rc;
+  { foreign_t rc;
 
     rc = pl_break1(ATOM_dquery_loop);
     restoreWakeup(&wstate);
@@ -276,14 +278,18 @@ currentBreakLevel(void)
 
 #define NOTRACE_TRACE 0x1
 #define NOTRACE_DEBUG 0x2
+#define NOTRACE_LCO   0x4
+#define NOTRACE_VMI   0x8
 
 static
 PRED_IMPL("$notrace", 2, notrace, PL_FA_NOTRACE)
 { PRED_LD
   int flags = 0;
 
-  if ( debugstatus.tracing   ) flags |= NOTRACE_TRACE;
-  if ( debugstatus.debugging ) flags |= NOTRACE_DEBUG;
+  if ( debugstatus.tracing   )              flags |= NOTRACE_TRACE;
+  if ( debugstatus.debugging )              flags |= NOTRACE_DEBUG;
+  if ( truePrologFlag(PLFLAG_LASTCALL) )    flags |= NOTRACE_LCO;
+  if ( truePrologFlag(PLFLAG_VMI_BUILTIN) ) flags |= NOTRACE_VMI;
 
   if ( PL_unify_integer(A1, flags) &&
        PL_unify_int64(A2, debugstatus.skiplevel) )
@@ -303,17 +309,33 @@ static
 PRED_IMPL("$restore_trace", 2, restoretrace, PL_FA_NOTRACE)
 { PRED_LD
   int flags;
-  int64_t depth;
+  int64_t depthi;
 
   if ( PL_get_integer_ex(A1, &flags) &&
-       PL_get_int64_ex(A2, &depth) )
-  { debugstatus.tracing   = !!(flags&NOTRACE_TRACE);
+       PL_get_int64_ex(A2, &depthi) )
+  { size_t depth;
+
+    if ( depthi == -1 )
+      depth = SKIP_VERY_DEEP;
+    else if ( depthi < 0 || depthi > SIZE_MAX )
+      return PL_representation_error("size_t");
+    else
+      depth = (size_t)depthi;
+
+    debugstatus.tracing   = !!(flags&NOTRACE_TRACE);
     debugstatus.debugging = !!(flags&NOTRACE_DEBUG);
     debugstatus.skiplevel = depth;
-    if ( debugstatus.debugging )
-      clearPrologRunMode(RUN_MODE_NORMAL);
+
+    if ( (flags&NOTRACE_LCO) )
+      setPrologFlagMask(PLFLAG_LASTCALL);
     else
-      setPrologRunMode(RUN_MODE_NORMAL);
+      clearPrologFlagMask(PLFLAG_LASTCALL);
+
+    if ( (flags&NOTRACE_VMI) )
+      setPrologFlagMask(PLFLAG_VMI_BUILTIN);
+    else
+      clearPrologFlagMask(PLFLAG_VMI_BUILTIN);
+
     updateAlerted(LD);
 
     return TRUE;
@@ -579,7 +601,7 @@ setAccessLevel(access_level_t accept)
 Cut (!) as called via the  meta-call  mechanism has no effect.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-word
+foreign_t
 pl_metacut(void)
 { succeed;
 }
@@ -684,17 +706,29 @@ check_data(DECL_LD Word p, chk_data *context)
 
 last_arg:
 
+#ifndef O_ATTVAR
+#define isAttVar(p) FALSE
+#endif
+
   while(isRef(*p))
   { assert(!is_marked(p));
-    p2 = unRef(*p);
+    if ( storage(*p) == STG_LOCAL ) /* PushPtr()/PopPtr() reference? */
+    { for(int i=0; i<LD->tmp.top; i++)
+      { if ( valTermRef(LD->tmp.h[i]) == p )
+	{ p2 = unRef(*p);
+	  goto deref_ok;
+	}
+      }
+      printk(context, "Reference to local stack");
+    } else
+    { p2 = unRef(*p);
+    }
+  deref_ok:
     DEBUG(CHK_HIGHER_ADDRESS,
 	  { if ( p2 > p )
-	    {
-#ifdef O_ATTVAR
-	      if ( !isAttVar(*p2) )
-#endif
-		if ( !gc_status.blocked )
-		  printk(context, "Reference to higher address");
+	    { if ( !isAttVar(*p2) &&
+		   !gc_status.blocked )
+		printk(context, "Reference to higher address");
 	    }
 	  });
     if ( p2 == p )
@@ -748,8 +782,6 @@ last_arg:
       printk(context, "Indirect at %p not on global stack", a);
     if ( storage(*p) != STG_GLOBAL )
       printk(context, "Indirect data not on global");
-    if ( isBignum(*p) )
-      return key+(word) valBignum(*p);
     if ( isFloat(*p) )
       return key+(word) valFloat(*p);
     if ( isString(*p) )
